// Generated by odyn-gen, do not edit by hand.

#include <boost/optional.hpp>

#include <vcsn/algos/fwd.hh>
#include <vcsn/dyn/automaton.hh>
#include <vcsn/dyn/context.hh>
#include <vcsn/dyn/types.hh>
#include <vcsn/dyn/value.hh>
#include <vcsn/misc/export.hh>

namespace vcsn
{
  namespace odyn LIBVCSN_API
  {
    using identities = vcsn::dyn::identities;
    using location = vcsn::dyn::location;

    struct automaton;
    struct context;
    struct expansion;
    struct expression;
    struct label;
    struct polynomial;
    struct weight;


    /// An object-oriented wrapper around a dyn::automaton.
    struct automaton
    {
    public:
      automaton(const vcsn::dyn::automaton& val)
        : val_{val}
      {}

      /// Direction for proper.
      using direction = ::vcsn::dyn::direction;

      /// Create an automaton from a file, or from a string.
      automaton(const std::string& data = "",
                const std::string& format = "default",
                const std::string& filename = "",
                bool strip = true);

      /// The accessible subautomaton of \a aut.
      auto accessible() const
        -> automaton;

      /// Sum of two automata.
      ///
      /// \param algo   how to compute the result
      ///   - "standard"    both \a lhs and \a rhs are standard,
      ///                   build a standard automaton.
      ///   - "general"     no requirement on \a lhs and and \a rhs.
      ///   - "auto"        "standard" if both automata are standard,
      ///                   "general" otherwise.
      auto add(const automaton& rhs, const std::string& algo = "auto") const
        -> automaton;

      /// An ambiguous word, or raise if there is none.
      auto ambiguous_word() const
        -> label;

      /// The coaccessible subautomaton of \a aut.
      auto coaccessible() const
        -> automaton;

      /// The codeterminized automaton.
      ///
      /// \param algo
      ///   - "boolean"     use efficient bitsets
      ///   - "weighted"    accept non Boolean automata (might not terminate)
      ///   - "auto"        "boolean" if the automaton is Boolean,
      ///                   "weighted" otherwise.
      /// \pre  the labelset of \a aut must be free.
      auto codeterminize(const std::string& algo = "auto") const
        -> automaton;

      /// The cominimized automaton.
      ///
      /// \param algo       the specific algorithm to use
      /// \pre  \a aut must be LAL.
      /// \pre  \a aut must be deterministic.
      auto cominimize(const std::string& algo = "auto") const
        -> automaton;

      /// Three-way comparison between automata.
      /// \returns negative if `lhs < rhs`
      ///          null     if `lhs = rhs`
      ///          positive if `lhs > rhs`
      auto compare(const automaton& rhs) const
        -> int;

      /// The complement of \a aut.
      ///
      /// \pre aut is lal
      /// \pre aut is Boolean
      /// \pre aut is deterministic
      /// \pre aut is complete
      auto complement() const
        -> automaton;

      /// A completed copy of \a aut.
      /// \pre aut is lal
      auto complete() const
        -> automaton;

      /// The automaton of a strongly connected component.
      ///
      /// \param num     the component number.
      auto component(unsigned num) const
        -> automaton;

      /// The composition of transducers \a lhs and \a rhs.
      ///
      /// \param lazy   whether to perform the computations on demand.
      auto compose(const automaton& rhs, bool lazy = false) const
        -> automaton;

      /// The condensation of \a aut such that each state is a strongly
      /// connected component.
      auto condense() const
        -> automaton;

      /// The conjugate of \a aut
      auto conjugate() const
        -> automaton;

      /// The conjunction (aka synchronized product) of automata.
      /// Performs the meet of the contexts.
      ///
      /// \param lazy   whether to perform the computations on demand.
      auto conjunction(const automaton& rhs, bool lazy = false) const
        -> automaton;

      /// The conjunction (aka synchronized product) of automata.
      /// Performs the meet of the contexts.
      ///
      /// \param lazy   whether to perform the computations on demand.
      static auto conjunction(const std::vector<automaton>& as, bool lazy = false)
        -> automaton;

      /// Repeated conjunction of \a aut with itself.
      auto conjunction(unsigned n) const
        -> automaton;

      /// The context of this automaton.
      auto context() const
        -> struct context;

      /// A copy of \a aut.
      auto copy() const
        -> automaton;

      /// A copy of \a aut converted to context \a ctx.
      auto copy(const struct context& ctx) const
        -> automaton;

      /// A co-standardized \a a.
      auto costandard() const
        -> automaton;

      /// The automaton with the delay of each state.
      auto delay_automaton() const
        -> automaton;

      /// The determinized automaton.
      ///
      /// \param algo
      ///     - "boolean"     use efficient bitsets
      ///     - "weighted"    accept non Boolean automata (might not terminate)
      ///     - "auto"        "boolean" if the automaton is Boolean,
      ///                     "weighted" otherwise.
      /// \pre  the labelset of \a aut must be free.
      auto determinize(const std::string& algo = "auto") const
        -> automaton;

      /// An automaton whose behavior is that of \a lhs on words not
      /// accepted by \a rhs.
      ///
      /// \pre \a rhs is Boolean.
      auto difference(const automaton& rhs) const
        -> automaton;

      /// The LAO automaton \a aut with state \a s removed.
      auto eliminate_state(int s = -1) const
        -> automaton;

      /// Evaluate \a l on \a aut.
      auto evaluate(const label& l) const
        -> weight;

      /// Evaluate \a p on \a aut.
      auto evaluate(const polynomial& p) const
        -> weight;

      /// Create a factor automaton from \a aut.
      auto factor() const
        -> automaton;

      /// The subautomaton based on \a aut, with only states in \a ss visible.
      auto filter(const std::vector<unsigned>& ss) const
        -> automaton;

      /// Focus on a specific tape of a tupleset automaton.
      auto focus(unsigned tape) const
        -> automaton;

      /// Whether the transducer has bounded lag.
      auto has_bounded_lag() const
        -> bool;

      /// Whether the automaton has a cycle with lightening weight.
      auto has_lightening_cycle() const
        -> bool;

      /// Whether the automaton has the twins property.
      auto has_twins_property() const
        -> bool;

      /// The infiltration of automata \a lhs and \a rhs.
      /// Performs the join of their types.
      auto infiltrate(const automaton& rhs) const
        -> automaton;

      /// The infiltration product of automata.
      /// Performs the join of their types.
      static auto infiltrate(const std::vector<automaton>& as)
        -> automaton;

      /// Split automaton on the incoming transition.
      ///
      /// In an insplit automaton, each state has either only
      /// spontaneous incoming transitions or only proper incoming
      /// transitions.
      ///
      /// \param lazy   whether to perform the computations on demand.
      ///
      /// \pre aut is lal.
      auto insplit(bool lazy = false) const
        -> automaton;

      /// Whether \a aut is accessible.
      auto is_accessible() const
        -> bool;

      /// Whether \a aut is ambiguous.
      ///
      /// \pre \a aut is LAL.
      auto is_ambiguous() const
        -> bool;

      /// Whether \a aut is coaccessible.
      auto is_coaccessible() const
        -> bool;

      /// Whether \a aut is codeterministic.
      ///
      /// \pre \a aut is LAL.
      auto is_codeterministic() const
        -> bool;

      /// Whether \a aut is complete.
      ///
      /// \pre \a aut is LAL.
      auto is_complete() const
        -> bool;

      /// Whether is costandard (unique final state, with weight one, no
      /// outcoming transition).
      auto is_costandard() const
        -> bool;

      /// Whether the automaton is cycle-ambiguous.
      auto is_cycle_ambiguous() const
        -> bool;

      /// Whether \a aut is deterministic.
      /// \pre \a aut is LAL.
      auto is_deterministic() const
        -> bool;

      /// Whether has no state.
      auto is_empty() const
        -> bool;

      /// Whether has no cycle of spontaneous transitions.
      auto is_eps_acyclic() const
        -> bool;

      /// Whether are the same automaton.
      auto is_equal(const automaton& rhs) const
        -> bool;

      /// Whether compute the same series.
      /// \pre The labelsets of lhs and rhs are free.
      /// \pre lhs and rhs are Boolean, or on Z, or on a field.
      auto is_equivalent(const automaton& rhs) const
        -> bool;

      /// Whether \a aut is free.
      auto is_free() const
        -> bool;

      /// Whether \a aut is functional.
      ///
      /// \pre \a aut is a transducer.
      auto is_functional() const
        -> bool;

      /// Whether there exists an isomorphism between the states of \a
      /// lhs and those of \a rhs.
      auto is_isomorphic(const automaton& rhs) const
        -> bool;

      /// Whether \a aut is letterized.
      auto is_letterized() const
        -> bool;

      /// Whether is normalized (in the Thompson sense), i.e., standard
      /// and co-standard.
      auto is_normalized() const
        -> bool;

      /// Whether the outgoing transitions of each state have increasing labels.
      auto is_out_sorted() const
        -> bool;

      /// Whether \a aut realizes a partial identity.
      ///
      /// \pre \a aut is a transducer.
      auto is_partial_identity() const
        -> bool;

      /// Whether has no spontaneous transition.
      auto is_proper() const
        -> bool;

      /// Whether has no spontaneous transition, and is letterized.
      auto is_realtime() const
        -> bool;

      /// Whether is standard (unique initial state, with weight one, no
      /// incoming transition).
      auto is_standard() const
        -> bool;

      /// Whether is synchronized.
      auto is_synchronized() const
        -> bool;

      /// Whether the word synchronizes aut.
      auto is_synchronized_by(const label& word) const
        -> bool;

      /// Whether is synchronizing.
      auto is_synchronizing() const
        -> bool;

      /// Whether has no useless state.
      auto is_trim() const
        -> bool;

      /// Whether has no useful state.
      auto is_useless() const
        -> bool;

      /// Whether automaton is valid (epsilon-cycles converge).
      auto is_valid() const
        -> bool;

      /// Left-division of two automata (lhs \ rhs).
      auto ldivide(const automaton& rhs) const
        -> automaton;

      /// Ordering between automata.
      auto less_than(const automaton& rhs) const
        -> bool;

      /// The automaton with letter transitions instead of words.
      auto letterize() const
        -> automaton;

      /// Lift some tapes of the transducer, or turn an automaton into a
      /// spontaneous automaton.
      ///
      /// Each `<k>l` transition is mapped to a `<<k>l>\\e` transition.
      ///
      /// \param tapes   the tapes to lift to weights
      /// \param ids     the identities of the expression
      auto lift(const std::vector<unsigned>& tapes = {}, identities ids = {}) const
        -> automaton;

      /// The approximated behavior of an automaton with smallest weights.
      ///
      /// \param num    number of words looked for.
      /// \param algo   the specific algorithm to use
      auto lightest(unsigned num = 1, const std::string& algo = "auto") const
        -> polynomial;

      /// The shortest path automaton using the given algorithm.
      auto lightest_automaton(unsigned num = 1, const std::string& algo = "auto") const
        -> automaton;

      /// The left-multiplication of an automaton with \a w as weight.
      ///
      /// \param algo   how to compute the result.
      ///   - "standard"    \a aut is standard,
      ///                   build a standard automaton.
      ///   - "general"     no requirement on \a aut,
      ///                   but add spontaneous transitions.
      ///   - "auto"        "standard" if \a aut is standard,
      ///                   "general" otherwise.
      auto lweight(const weight& w, const std::string& algo = "auto") const
        -> automaton;

      /// Read an automaton from a string.
      /// \param data     the input string.
      /// \param format   its format.
      /// \param strip    whether to return a stripped automaton,
      ///                or a named automaton.
      auto make_automaton(const std::string& data, const std::string& format = "default", bool strip = true)
        -> automaton;

      /// The minimized automaton.
      ///
      /// \param algo       the specific algorithm to use
      /// \pre  \a aut must be LAL.
      /// \pre  \a aut must be deterministic.
      auto minimize(const std::string& algo = "auto") const
        -> automaton;

      /// Multiply (concatenate) two automata.
      ///
      /// \param algo   how to compute the result
      ///   - "standard"    both \a lhs and \a rhs are standard,
      ///                   build a standard automaton.
      ///   - "general"     no requirement on \a lhs and and \a rhs,
      ///                   but add spontaneous transitions.
      ///   - "auto"        "standard" if both automata are standard,
      ///                   "general" otherwise.
      auto multiply(const automaton& rhs, const std::string& algo = "auto") const
        -> automaton;

      /// Repeated multiplication (concatenation) of an automaton with itself.
      ///
      /// \param min   the minimum number.  If -1, denotes 0.
      /// \param max   the maximum number.
      ///             If -1, denotes infinity, using star.
      ///             If -2, denotes the same value as min.
      /// \param algo   how to compute the result
      ///   - "standard"    \a aut is standard,
      ///                   build a standard automaton.
      ///   - "general"     no requirement on \a aut,
      ///                   but add spontaneous transitions.
      ///   - "auto"        "standard" if \a aut is standard,
      ///                   "general" otherwise.
      auto multiply(int min, int max = -2, const std::string& algo = "auto") const
        -> automaton;

      /// Normalize automaton \a aut.
      auto normalize() const
        -> automaton;

      /// The number of strongly connected components.
      auto num_components() const
        -> std::size_t;

      /// Build the pair automaton of the given automaton
      auto pair(bool keep_initials = false) const
        -> automaton;

      /// Create a partial identity transducer from \a aut.
      auto partial_identity() const
        -> automaton;

      /// Create a prefix automaton from \a aut.
      auto prefix() const
        -> automaton;

      /// Keep a single tape from a multiple-tape automaton.
      auto project(unsigned tape) const
        -> automaton;

      /// An equivalent automaton without spontaneous transitions.
      ///
      /// \param dir     the direction of the closure
      /// \param prune   whether to delete states that become inaccessible
      /// \param algo    how elimination is performed
      ///     - "auto"       same as "inplace"
      ///     - "default"    same as "inplace"
      ///     - "inplace"    eliminate in place
      ///     - "separate"   first separate the automaton into spontaneous and
      ///                    proper parts
      ///     - "distance"   compute all-pairs distances
      auto proper(direction dir = direction::backward, bool prune = true, const std::string& algo = "auto") const
        -> automaton;

      /// The weight pushing automaton of \a aut.
      auto push_weights() const
        -> automaton;

      /// Right-division of two automata (lhs / rhs).
      auto rdivide(const automaton& rhs) const
        -> automaton;

      /// Read an automaton from a stream.
      /// \param is       the input stream.
      /// \param format   its format ("auto", "daut", "default", "dot",
      ///                "efsm", "fado", "grail").  "default" means
      ///                "auto": try to guess the format.
      /// \param strip    whether to return a stripped automaton,
      ///                or a named automaton.
      /// \param loc      the location before the first character in \a is
      auto read_automaton(std::istream& is, const std::string& format = "default", bool strip = true, const location& loc = location{})
        -> automaton;

      /// Create the realtime automaton (letterized and proper).
      auto realtime() const
        -> automaton;

      /// Reduce \a aut.
      auto reduce() const
        -> automaton;

      /// The right-mult automaton with \a w as weight.
      ///
      /// \param algo   how to compute the result.
      ///              In most cases, "standard" and "general" have
      ///              equal results.  The only difference in when
      ///              when \a w is null, in which case "general"
      ///              produces an empty automaton, while "standard"
      ///              produces an automaton with a single state, which
      ///              is initial (and, of course, no final states).
      ///   - "standard"    \a aut is standard,
      ///                   build a standard automaton.
      ///   - "general"     no requirement on \a aut,
      ///                   but add spontaneous transitions.
      ///   - "auto"        "standard" if \a aut is standard,
      ///                   "general" otherwise.
      auto rweight(const weight& w, const std::string& algo = "auto") const
        -> automaton;

      /// Build the SCC automaton whose states are labeled with number
      /// of the strongly-connected component they belong to.
      ///
      /// \param algo    the specific algorithm to use.
      auto scc(const std::string& algo = "auto") const
        -> automaton;

      /// The approximated behavior of an automaton.
      ///
      /// \param num    number of words looked for.
      /// \param len    maximum length of words looked for.
      auto shortest(boost::optional<unsigned> num = {}, boost::optional<unsigned> len = {}) const
        -> polynomial;

      /// The shuffle product of automata \a lhs and \a rhs.
      /// Performs the join of their type.
      auto shuffle(const automaton& rhs) const
        -> automaton;

      /// The shuffle product of automata.
      /// Performs the join of their types.
      static auto shuffle(const std::vector<automaton>& as)
        -> automaton;

      /// A copy of \a a with normalized state numbers.
      auto sort() const
        -> automaton;

      /// A standardized \a a.
      auto standard() const
        -> automaton;

      /// Star of an automaton.
      ///
      /// \param algo   how to compute the result.
      ///   - "standard"    \a aut is standard,
      ///                   build a standard automaton.
      ///   - "general"     no requirement on \a aut,
      ///                   but add spontaneous transitions.
      ///   - "auto"        "standard" if \a aut is standard,
      ///                   "general" otherwise.
      auto star(const std::string& algo = "auto") const
        -> automaton;

      /// The automaton in \a a with its metadata layers removed.
      auto strip() const
        -> automaton;

      /// Create a subword automaton from \a aut.
      ///
      /// For each non-spontaneous transition, add a spontaneous
      /// transition with same source, destination, and weight.
      auto subword() const
        -> automaton;

      /// Create a suffix automaton from \a aut.
      auto suffix() const
        -> automaton;

      /// Create a synchronized transducer from \a aut.
      auto synchronize() const
        -> automaton;

      /// A synchronizing word, or raise if there is none.
      auto synchronizing_word(const std::string& algo = "greedy") const
        -> label;

      /// An expression denoting the language of \a aut.
      ///
      /// \param ids    the identities to use for the expression
      /// \param algo   the state selection strategy
      auto to_expression(identities ids = {}, const std::string& algo = "auto") const
        -> expression;

      /// Transpose \a aut.
      auto transpose()
        -> automaton;

      /// The trim subautomaton of \a aut.
      auto trim() const
        -> automaton;

      /// Build a two-tape automaton from two automata.
      auto tuple(const automaton& rhs) const
        -> automaton;

      /// Build a k-tape automaton from k automata.
      static auto tuple(const std::vector<automaton>& as)
        -> automaton;

      /// The implementation type of a.
      auto type() const
        -> std::string;

      /// The universal automaton of \a aut.
      auto universal() const
        -> automaton;

      /// Compute weight of the series
      auto weight_series() const
        -> weight;

      /// The wrapped dyn::automaton.
      vcsn::dyn::automaton val_;
    };

    /// An object-oriented wrapper around a dyn::context.
    struct context
    {
    public:
      context(const vcsn::dyn::context& val)
        : val_{val}
      {}

      context();

      context(const std::string& ctx);

      explicit operator bool() const;

      automaton cotrie(const std::string& data,
                       const std::string& format,
                       const std::string& filename) const;

      automaton trie(const std::string& data,
                     const std::string& format,
                     const std::string& filename) const;

      /// Produce a Černý automaton of \a num_states states.
      auto cerny(unsigned num_states) const
        -> automaton;

      /// The composition of two contexts.
      auto compose(const struct context& rhs) const
        -> struct context;

      /// A reversed trie-like automaton (multiple initial states,
      /// single final state) automaton to accept the series contained
      /// in \a is.
      ///
      /// \param is       the input stream.
      /// \param format   the format of the file: "words" or "monomials".
      auto cotrie(std::istream& is, const std::string& format = "default") const
        -> automaton;

      /// A simple NFA for (a+b)*a(a+b)^n.
      auto de_bruijn(unsigned n) const
        -> automaton;

      /// An automaton which accepts a word n representing a number in
      /// base b iff k|n.
      auto divkbaseb(unsigned divisor, unsigned base) const
        -> automaton;

      /// The double_ring automaton with \a n states and \a f the list of finals.
      auto double_ring(unsigned n, const std::vector<unsigned>& f) const
        -> automaton;

      /// The expression for `\\e`.
      auto expression_one(identities ids) const
        -> expression;

      /// The expression for `\\z`.
      auto expression_zero(identities ids) const
        -> expression;

      /// The join between two contexts, i.e., their lowest common supertype.
      auto join(const struct context& rhs) const
        -> struct context;

      /// The ladybird automaton with \a n states.
      auto ladybird(unsigned n) const
        -> automaton;

      /// The transducer representing the Levenshtein distance.
      auto levenshtein() const
        -> automaton;

      /// Build a context from its name.
      auto make_context(const std::string& name)
        -> struct context;

      /// Read an expression from a stream.
      ///
      /// \param s        the input string.
      /// \param ids      the identities to apply.
      auto make_expression(const std::string& s, identities ids = {}, const std::string& format = "default") const
        -> expression;

      /// Build a label from a string.
      auto make_label(const std::string& s, const std::string& format = "default") const
        -> label;

      /// Build a polynomial from a string.
      auto make_polynomial(const std::string& s) const
        -> polynomial;

      /// Build a weight from a string.
      auto make_weight(const std::string& s) const
        -> weight;

      /// Build a word from a string.
      auto make_word(const std::string& s, const std::string& format = "default") const
        -> label;

      /// The context for words.
      auto make_word_context() const
        -> struct context;

      /// The number of input tapes.
      /// \returns 0 if the context is not tupleset.
      auto num_tapes() const
        -> std::size_t;

      /// Keep a single tape from a multiple-tape context.
      auto project(unsigned tape) const
        -> struct context;

      /// A transducer that computes the quotient of the division
      /// of a word n by k in base b iff k|n.
      auto quotkbaseb(unsigned divisor, unsigned base) const
        -> automaton;

      /// A random automaton.
      ///
      /// \param num_states
      ///    The number of states wanted in the automata
      ///    (>0).  All states will be connected, and there will be no dead
      ///    state.  However, some might not be coaccessible.
      /// \param density
      ///    The density of the automata.  This is the probability
      ///    (between 0.0 and 1.0), to add a transition between two
      ///    states.  All states have at least one outgoing transition,
      ///    so \a d is considered only when adding the remaining transition.
      ///    A density of 1 means all states will be connected to each other.
      /// \param num_initial
      ///    The number of initial states wanted (0 <= num_initial <= num_states)
      /// \param num_final
      ///    The number of final states wanted (0 <= num_final <= num_states)
      /// \param max_labels
      ///    The maximum number of labels per transition.  Defaults to the
      ///    number of generators.
      /// \param loop_chance
      ///    The probability (between 0.0 and 1.0) for each state to have
      ///    a loop.
      /// \param weights
      ///    The specification string (following the format of random_weight) used
      ///    to generate weights on each transitions.
      auto random_automaton(unsigned num_states, float density = 0.1, unsigned num_initial = 1, unsigned num_final = 1, boost::optional<unsigned> max_labels = {}, float loop_chance = 0.0, const std::string& weights = "") const
        -> automaton;

      /// A random deterministic automaton.
      ///
      /// \param num_states
      ///    The number of states wanted in the automata (>0).
      auto random_automaton_deterministic(unsigned num_states) const
        -> automaton;

      /// A random expression.
      ///
      /// \param param
      ///    A specification of the wanted operators and their relative
      ///    probability.  May also contain the maximum number of
      ///    symbols of the expression with key `length`.  For instance:
      ///    "l, +=1,.=2,{T}=0.5,length=6".
      /// \param ids
      ///    The identities to use for the expression.
      auto random_expression(const std::string& param = "default", identities ids = {}) const
        -> expression;

      /// A random label.
      ///
      /// \param param
      ///    A string containing paramaters for the generation.
      auto random_label(const std::string& param) const
        -> label;

      /// A random weight.
      ///
      /// \param param
      ///    A string containing paramaters for the generation such as the
      ///    min and max, or an element you want with specific probability.
      ///    For instance: "1=0.2, 10=0.3, min=0, max=20".
      auto random_weight(const std::string& param) const
        -> weight;

      /// Read an expression from a stream.
      ///
      /// \param ids      the identities to apply.
      /// \param is       the input stream.
      /// \param format   the expression's format.
      /// \param loc      the location before the first character in \a is
      auto read_expression(identities ids, std::istream& is, const std::string& format = "default", const location& loc = location{}) const
        -> expression;

      /// Read a label from a stream.
      ///
      /// \param is       the input stream.
      /// \param format   the label's format.
      auto read_label(std::istream& is, const std::string& format = "default") const
        -> label;

      /// Read a polynomial from a stream
      ///
      /// \param is     the stream to parse.
      auto read_polynomial(std::istream& is) const
        -> polynomial;

      /// Read a weight from a stream.
      ///
      /// \param is     the input stream.
      auto read_weight(std::istream& is) const
        -> weight;

      /// An expression denoting the label of \a l.
      auto to_expression(identities ids, const label& l) const
        -> expression;

      /// An expression matching one letter in a letter class.
      ///
      /// \param ids
      ///   The identities to use for the expression
      /// \param letters
      ///   The letter class as a set of ranges.
      /// \param accept
      ///   Whether to accept these characters ([abc]) as opposed
      ///   to refusing them ([^abc]).
      auto to_expression(identities ids, const vcsn::letter_class_t& letters, bool accept = true) const
        -> expression;

      /// A trie-like automaton (single initial state, multiple final
      /// states) to accept the series contained in \a is.
      ///
      /// \param is       the input stream.
      /// \param format   the format of the file: "words" or "monomials".
      auto trie(std::istream& is, const std::string& format = "default") const
        -> automaton;

      /// Tuple two contexts.
      auto tuple(const struct context& rhs) const
        -> struct context;

      /// A k-tape context from k contexts.
      static auto tuple(const std::vector<struct context>& ctxs)
        -> struct context;

      /// The Brzozowski universal witness.
      auto u(unsigned n) const
        -> automaton;

      /// The weight one.
      auto weight_one() const
        -> weight;

      /// The weight zero.
      auto weight_zero() const
        -> weight;

      /// The wrapped dyn::context.
      vcsn::dyn::context val_;
    };

    /// An object-oriented wrapper around a dyn::expansion.
    struct expansion
    {
    public:
      expansion(const vcsn::dyn::expansion& val)
        : val_{val}
      {}

      /// Sum of two expansions.
      auto add(const expansion& rhs) const
        -> expansion;

      /// Complement this expansion.
      auto complement() const
        -> expansion;

      /// The composition of two expansions.
      auto compose(const expansion& rhs) const
        -> expansion;

      /// The Hadamard product of expansions \a lhs and \a rhs.
      auto conjunction(const expansion& rhs) const
        -> expansion;

      /// The context of this expansion.
      auto context() const
        -> struct context;

      /// Denormalize expansion \a x.
      auto denormalize() const
        -> expansion;

      /// The determinized expansion.
      auto determinize() const
        -> expansion;

      /// Whether has no polynomial for the empty word.
      auto is_normal() const
        -> bool;

      /// Left-division of two expansions (lhs \ rhs).
      auto ldivide(const expansion& rhs) const
        -> expansion;

      /// The left-multiplication of an expansion with \a w as weight.
      auto lweight(const weight& w) const
        -> expansion;

      /// Normalize expansion \a x.
      auto normalize() const
        -> expansion;

      /// Keep a single tape from a multiple-tape expansion.
      auto project(unsigned tape) const
        -> expansion;

      /// The right-multiplication of an expansion with \a w as weight.
      auto rweight(const weight& w) const
        -> expansion;

      /// Projection as a expression.
      auto to_expression() const
        -> expression;

      /// Build a k-tape expansion from k expansions.
      static auto tuple(const std::vector<expansion>& rs)
        -> expansion;

      /// The wrapped dyn::expansion.
      vcsn::dyn::expansion val_;
    };

    /// An object-oriented wrapper around a dyn::expression.
    struct expression
    {
    public:
      expression(const vcsn::dyn::expression& val)
        : val_{val}
      {}

      expression(const struct context& ctx, const std::string& r,
                 identities ids = {});

      /// Convert \a this to \a ctx, using \a ids.
      expression as(const struct context& ctx = {},
                    identities ids = "default") const;

      /// Sum of two expressions.
      auto add(const expression& rhs) const
        -> expression;

      /// Three-way comparison between expressions.
      /// \returns negative if `lhs < rhs`
      ///          null     if `lhs = rhs`
      ///          positive if `lhs > rhs`
      auto compare(const expression& rhs) const
        -> int;

      /// Add the complement operator to \a r.
      auto complement() const
        -> expression;

      /// The composition of two expressions.
      auto compose(const expression& rhs) const
        -> expression;

      /// Concatenate two expressions.
      ///
      /// Do not use this routine, see multiply instead.
      auto concatenate(const expression& rhs) const
        -> expression;

      /// The Hadamard product of expressions \a lhs and \a rhs.
      auto conjunction(const expression& rhs) const
        -> expression;

      /// The weight associated to the empty word in \a e.
      auto constant_term() const
        -> weight;

      /// The context of this expression.
      auto context() const
        -> struct context;

      /// Copy of \a exp, but built with \a ctx.
      auto copy(const struct context& ctx, identities ids) const
        -> expression;

      /// Derive \a exp with respect to \a s.
      ///
      /// \param breaking   whether to split the result
      /// \pre  \a exp must be LAL.
      auto derivation(const label& lbl, bool breaking = false) const
        -> polynomial;

      /// The derived-term automaton of \a exp.
      ///
      /// \param algo    how the derived terms are computed:
      ///    - "auto"                  alias for "expansion"
      ///    - "derivation"            compute by derivation
      ///    - "breaking_derivation"   compute by breaking derivation
      ///    - "expansion"             compute by expansion
      ///    - "breaking_expansion"    compute by breaking expansion
      auto derived_term(const std::string& algo = "auto") const
        -> automaton;

      /// Words accepted by \a lhs, but not by \a rhs.
      auto difference(const expression& rhs) const
        -> expression;

      /// Distribute product over addition recursively under the starred
      /// subexpressions and group the equal monomials.
      auto expand() const
        -> expression;

      /// The identities of expression \a exp.
      auto identities_of() const
        -> identities;

      /// The inductive translation of \a e in an automaton.
      ///
      /// \param algo   how to compute the result
      ///   - "standard"    build a standard automaton.
      ///   - "general"     no requirement on the result
      ///   - "auto"        same as "standard"
      auto inductive(const std::string& algo = "auto") const
        -> automaton;

      /// The infiltration product of expressions \a lhs and \a rhs.
      /// Performs the join of their type.
      auto infiltrate(const expression& rhs) const
        -> expression;

      /// Whether are the same expression.
      auto is_equal(const expression& rhs) const
        -> bool;

      /// Whether denote the same series.
      /// \pre The labelsets of lhs and rhs are free.
      /// \pre lhs and rhs are Boolean, or on Z, or on a field.
      auto is_equivalent(const expression& rhs) const
        -> bool;

      /// Whether rational expression is valid (all the starred
      /// sub-expressions are starrable).
      auto is_valid() const
        -> bool;

      /// Left-division of two expressions (lhs \ rhs).
      auto ldivide(const expression& rhs) const
        -> expression;

      /// Ordering between expressions.
      auto less_than(const expression& rhs) const
        -> bool;

      /// The lifted LAO rational expression from \a e.
      auto lift() const
        -> expression;

      /// The left-multiplication of an expression with \a w as weight.
      auto lweight(const weight& w) const
        -> expression;

      /// Multiply (concatenate) two expressions.
      auto multiply(const expression& rhs) const
        -> expression;

      /// Repeated multiplication (concatenation) of an expression with itself.
      ///
      /// \param min   the minimum number.  If -1, denotes 0.
      /// \param max   the maximum number.
      ///             If -1, denotes infinity, using star.
      ///             If -2, denotes the same value as min.
      auto multiply(int min, int max = -2) const
        -> expression;

      /// Name an expression.
      auto name(const std::string& name) const
        -> expression;

      /// Create a partial-identity two-tape expression from \a e.
      auto partial_identity() const
        -> expression;

      /// Keep a single tape from a multiple-tape expression.
      auto project(unsigned tape) const
        -> expression;

      /// Right-division of two expressions (lhs / rhs).
      auto rdivide(const expression& rhs) const
        -> expression;

      /// The right-multiplication of an expression with \a w as weight.
      auto rweight(const weight& w) const
        -> expression;

      /// The shuffle product of expressions \a lhs and \a rhs.
      /// Performs the join of their type.
      auto shuffle(const expression& rhs) const
        -> expression;

      /// Break \a exp.
      auto split() const
        -> polynomial;

      /// The standard automaton of \a e.
      auto standard() const
        -> automaton;

      /// Star height of an expression.
      auto star_height() const
        -> unsigned;

      /// A normalized form where star is applied only to proper
      /// expressions.
      ///
      /// \pre \a e is Boolean.
      auto star_normal_form() const
        -> expression;

      /// The Thompson automaton of \a e.
      auto thompson() const
        -> automaton;

      /// An automaton denoting the language of \a exp.
      /// \param algo   the conversion algorithm.
      ///   - "auto": currently equivalent to "expansion", eventually should
      ///      mix "standard" for basic expressions (faster) and "expansion"
      ///      otherwise (more general).
      ///   - "derivation": use derivation-based expression.derived_term,
      ///      stripped.
      ///   - "derived_term": same as "expansion"
      ///   - "expansion": use expansion-based expression.derived_term, stripped.
      ///   - "inductive": use expression.inductive
      ///   - "standard": use expression.standard.
      ///   - "thompson": use expression.thompson.
      ///   - "zpc": use expression.zpc.
      ///   - "zpc,compact": use expression.zpc, "compact" version.
      auto to_automaton(const std::string& algo = "auto") const
        -> automaton;

      /// First order development of a \a exp.
      auto to_expansion() const
        -> expansion;

      /// Transpose \a e.
      auto transpose() const
        -> expression;

      /// Add the transposition operator to \a r.
      auto transposition() const
        -> expression;

      /// Build a two-tape expression from two expressions.
      auto tuple(const expression& rhs) const
        -> expression;

      /// Build a k-tape expression from k expressions.
      static auto tuple(const std::vector<expression>& rs)
        -> expression;

      /// The ZPC automaton of \a exp.
      ///
      /// \param algo    the specific algorithm to use.
      ///               It can be "regular"/"auto" or the variant "compact".
      auto zpc(const std::string& algo = "auto") const
        -> automaton;

      /// The wrapped dyn::expression.
      vcsn::dyn::expression val_;
    };

    /// An object-oriented wrapper around a dyn::label.
    struct label
    {
    public:
      label(const vcsn::dyn::label& val)
        : val_{val}
      {}

      label(const context& ctx, const std::string& s);

      /// Three-way comparison between labels.
      /// \returns negative if `lhs < rhs`
      ///          null     if `lhs = rhs`
      ///          positive if `lhs > rhs`
      auto compare(const label& rhs) const
        -> int;

      /// The composition of two labels.
      auto compose(const label& rhs) const
        -> label;

      /// Left-division of two labels (lhs \ rhs).
      auto ldivide(const label& rhs) const
        -> label;

      /// Multiply (concatenate) two labels.
      auto multiply(const label& rhs) const
        -> label;

      /// Repeated multiplication of a label with itself.
      ///
      /// \param num   the exponent.
      auto multiply(int num) const
        -> label;

      /// Keep a single tape from a multiple-tape label.
      auto project(unsigned tape) const
        -> label;

      /// Right-division of two labels (lhs / rhs).
      auto rdivide(const label& rhs) const
        -> label;

      /// The wrapped dyn::label.
      vcsn::dyn::label val_;
    };

    /// An object-oriented wrapper around a dyn::polynomial.
    struct polynomial
    {
    public:
      polynomial(const vcsn::dyn::polynomial& val)
        : val_{val}
      {}

      polynomial(const struct context& ctx, const std::string& s);

      /// Sum of two polynomials.
      auto add(const polynomial& rhs) const
        -> polynomial;

      /// Three-way comparison between polynomials.
      /// \returns negative if `lhs < rhs`
      ///          null     if `lhs = rhs`
      ///          positive if `lhs > rhs`
      auto compare(const polynomial& rhs) const
        -> int;

      /// The composition of two polynomials.
      auto compose(const polynomial& rhs) const
        -> polynomial;

      /// The Hadamard product of polynomials \a lhs and \a rhs.
      auto conjunction(const polynomial& rhs) const
        -> polynomial;

      /// The context of this polynomial.
      auto context() const
        -> struct context;

      /// A reversed trie-like automaton (multiple initial states,
      /// single final state) automaton to accept \a p.
      auto cotrie() const
        -> automaton;

      /// The infiltration product of polynomial \a lhs and \a rhs.
      auto infiltrate(const polynomial& rhs) const
        -> polynomial;

      /// Left-division of two polynomials (lhs \ rhs).
      auto ldivide(const polynomial& rhs) const
        -> polynomial;

      /// Left greatest common divisor of two polynomials (lhs \ rhs).
      auto lgcd(const polynomial& rhs) const
        -> polynomial;

      /// The left-multiplication of a polynomial with \a w as weight.
      auto lweight(const weight& w) const
        -> polynomial;

      /// Multiply two polynomials.
      auto multiply(const polynomial& rhs) const
        -> polynomial;

      /// Keep a single tape from a multiple-tape polynomial.
      auto project(unsigned tape) const
        -> polynomial;

      /// The right-multiplication of a polynomial with \a w as weight.
      auto rweight(const weight& w) const
        -> polynomial;

      /// The shuffle product of polynomial \a lhs and \a rhs.
      auto shuffle(const polynomial& rhs) const
        -> polynomial;

      /// Break all the expressions in \a p.
      auto split() const
        -> polynomial;

      /// A trie-like automaton (single initial state, multiple final
      /// states) to accept \a p.
      auto trie() const
        -> automaton;

      /// Build a k-tape polynomial from k polynomials.
      static auto tuple(const std::vector<polynomial>& ps)
        -> polynomial;

      /// The wrapped dyn::polynomial.
      vcsn::dyn::polynomial val_;
    };

    /// An object-oriented wrapper around a dyn::weight.
    struct weight
    {
    public:
      weight(const vcsn::dyn::weight& val)
        : val_{val}
      {}

      weight(const context& ctx, const std::string& s);

      /// Sum of two weights.
      auto add(const weight& rhs) const
        -> weight;

      /// Three-way comparison between weights.
      /// \returns negative if `lhs < rhs`
      ///          null     if `lhs = rhs`
      ///          positive if `lhs > rhs`
      auto compare(const weight& rhs) const
        -> int;

      /// Left-division of two weights (lhs \ rhs).
      auto ldivide(const weight& rhs) const
        -> weight;

      /// Multiply two weights.
      auto multiply(const weight& rhs) const
        -> weight;

      /// Repeated multiplication of a weight with itself.
      ///
      /// \param min   the minimum number.  If -1, denotes 0.
      /// \param max   the maximum number.
      ///             If -1, denotes infinity, using star.
      ///             If -2, denotes the same value as min.
      auto multiply(int min, int max = -2) const
        -> weight;

      /// Right-division of two weights (lhs / rhs).
      auto rdivide(const weight& rhs) const
        -> weight;

      /// The wrapped dyn::weight.
      vcsn::dyn::weight val_;
    };
  }
}
